/**********************************************************************/
/*
     FILE: cwfatt.c
  PURPOSE: To print an attribute value from a CoastWatch format file.
   AUTHOR: Peter Hollemans
     DATE: 05/06/1998
  CHANGES: 01/09/1998, PFH, changed format of attribute output list

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include "cwf.h"

/* defined constants */

#define UMESSAGE "\
usage: cwfatt [-a attribute] input\n\
where: -a attribute     The attribute to print.  If not specified, a list\n\
                        of valid attribute names is printed.\n\
       input            The CoastWatch dataset to read.\n"

#define NARG	1

/* defined macros */

#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 

#define STR(e) cw_strerror ((e))

static int cwerr;
#define CHK(r,v) { \
  if ((cwerr = (r)) != CW_NOERR) \
    XIT (STR (cwerr), (v), cwerr); \
} 

/**********************************************************************/

void err (const char *m, const char *v) {
  fprintf (stderr, "cwfatt: %s: %s\n", v, m);
} /* err */

/**********************************************************************/

void usage (int ecode) {
  fprintf (stderr, UMESSAGE);
  exit (ecode);
} /* usage */

/**********************************************************************/

int main (int argc, char *argv[]) {

short short_val;
float float_val;
char text_val[50];
int cwid, varid, natts;
cw_type att_xtype;
char att_name[CW_MAX_NAME];
int list = 1;
int i;
char cbuf[32];

  argc--, argv++;				/* parse command line */
  if (argc < NARG)
    usage (-1);
  while (argv[0][0] == '-') {
    switch (argv[0][1]) {
      case 'a':
        strcpy (att_name, argv[1]);
        list = 0;
        argc-=2, argv+=2; 
        break;
      default: 
        err ("invalid option", argv[0]); 
        usage (-2);
    } /* switch */
    if (argc < NARG)
      usage (-3);
  } /* while */

						/* open file */
  CHK (cw_open (argv[0], CW_NOWRITE, &cwid), argv[0]);

  varid = 0;					/* get variable info */
  CHK (cw_inq_var (cwid, varid, NULL, NULL, NULL, NULL, &natts), "variable 0");

  if (list) {					/* list attributes */
    for (i = 0; i < natts; i++) {
      sprintf (cbuf, "attribute %d", i);
      CHK (cw_inq_attname (cwid, varid, i, att_name), cbuf);
      CHK (cw_inq_att (cwid, varid, att_name, &att_xtype, NULL), att_name);
      printf ("%s\n", att_name);
    } /* for */
  } /* if */
  else {
    CHK (cw_inq_att (cwid, varid, att_name, &att_xtype, NULL), att_name);
    switch (att_xtype) {
      case CW_CHAR:
        if (cw_get_att_text (cwid, varid, att_name, text_val) != CW_NOERR)
          strcpy (text_val, "undefined");
        printf ("%s\n", text_val);
        break;
      case CW_SHORT:
        CHK (cw_get_att_short (cwid, varid, att_name, &short_val), att_name);
        printf ("%hd\n", short_val);
        break;
      case CW_FLOAT:
        CHK (cw_get_att_float (cwid, varid, att_name, &float_val), 
          att_name);
        printf ("%#.6g\n", float_val);
        break;
    } /* switch */
  } /* else */

  CHK (cw_close (cwid), argv[0]);		/* finish up */
  return (CW_NOERR);

} /* main */

/**********************************************************************/
