/**********************************************************************/
/*
     FILE: cwflmask.c
  PURPOSE: To generate a CoastWatch file land mask and apply the mask to
           the data.
   AUTHOR: Peter Hollemans
     DATE: 1999/07/29
  CHANGES: n/a

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998-1999, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cwf.h"
#include "polyfill.h"
#include "morph.h"

/* defined constants */

#define UMESSAGE "\
usage: cwflmask [-o r,c] [-r] [-e n] input output\n\
where: -o r,c           Specifies a 1-relative row and column point for\n\
                        open ocean, default is to assume that the lower-left\n\
                        corner is an ocean point.\n\
       -r               Attempt to generate a recursive land mask - may\n\
                        not work well for ambiguous and/or multiply\n\
                        connected coastlines.\n\
       -e n             Optionally extend the land mask out from the\n\
                        coast by n pixels.\n\
       input            The CoastWatch dataset to mask.\n\
       output           The CoastWatch dataset to write.  The output\n\
                        values will be altered so that land pixels are\n\
                        flagged as invalid data, and graphics plane 1\n\
                        will contain the land mask.\n"
#define NARG	2
#define BUFSIZE	512
#define OCEAN	0
#define LAND	1

/* defined macros */

#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 

#define STR(e) cw_strerror ((e))

static int cwerr;
#define CHK(r,v) { \
  if ((cwerr = (r)) != CW_NOERR) \
    XIT (STR (cwerr), (v), cwerr); \
} 

/**********************************************************************/

void err (			/*** PRINT ERROR MESSAGE ***/
  const char *m,		/* error message string */
  const char *v			/* variable that message applies to */
) {

  fprintf (stderr, "cwflmask: %s: %s\n", v, m);

} /* err */

/**********************************************************************/

void usage (			/*** PRINT USAGE MESSAGE ***/
  int ecode			/* exit status code */
) {

  fprintf (stderr, UMESSAGE);
  exit (ecode);

} /* usage */

/**********************************************************************/

int cp (			/*** COPIES FILE A TO FILE B ***/
  char *a,			/* input file name */
  char *b			/* output file name */
) {				/* return: 0 on success, 1 on failure */

size_t num_read;
unsigned char buf[BUFSIZE];
FILE *afp, *bfp;

  /* Open input and output files */
  /* --------------------------- */
  afp = fopen (a, "rb");
  bfp = fopen (b, "wb");
  if (afp == NULL || bfp == NULL)
    return (1);

  /* Read and write data */
  /* ------------------- */
  num_read = fread ((void *) buf, 1, BUFSIZE, afp);
  while (num_read != 0) {
    if (fwrite ((void *) buf, 1, num_read, bfp) < num_read)
      return (1);
    num_read = fread ((void *) buf, 1, BUFSIZE, afp);
  } /* while */

  /* Close files and return */
  /* ---------------------- */
  fclose (afp);
  fclose (bfp);
  return (0);

} /* cp */

/**********************************************************************/

int main (int argc, char *argv[]) {

size_t ocean[2];
int guess_ocean = 1;
int recursive = 0;
int extend = 0;
char *ifname, *ofname;
int cwid, varid, gvarid;
int dimid[2];
size_t rows, cols, dvals, i, e;
cw_type xtype;
char var_name[CW_MAX_NAME];
float *fp;
unsigned char *ucp, *landmask;
size_t start[2], count[2];
int row, col;
int len;
unsigned char se[] = {
  1,1,1,
  1,1,1,
  1,1,1
};

  /* Parse command line */
  /* ------------------ */
  argc--, argv++;
  if (argc < NARG)
    usage (-1);
  while (argv[0][0] == '-') {
    switch (argv[0][1]) {

      case 'o':
        sscanf (argv[1], "%d,%d", ocean, ocean+1);
        guess_ocean = 0;
        argc-=2, argv+=2; 
        break;

      case 'r':
        recursive = 1;
        argc--, argv++;
        break;

      case 'e':
        extend = atoi (argv[1]);
        argc-=2, argv+=2; 
        break;

     default: 
        err ("invalid option", argv[0]); 
        usage (-2);

    } /* switch */
    if (argc < NARG)
      usage (-2);
  } /* while */

  ifname = argv[0];
  ofname = argv[1];

  /* Create output file */
  /* ------------------ */
  if (cp (ifname, ofname) != 0)
    XIT ("file copy failed", "cp", -3);
  CHK (cw_open (ofname, CW_WRITE, &cwid), ofname);

  /* Check for graphics planes */
  /* ------------------------- */
  if (cw_inq_varid (cwid, "graphics", &gvarid) != CW_NOERR)
    XIT ("no graphics planes found", "error", -4);

  /* Get variable dimensions */
  /* ----------------------- */
  varid = 0;    
  CHK (cw_inq_var (cwid, varid, var_name, &xtype, NULL, dimid, NULL),
    "variable 0");
  CHK (cw_inq_dim (cwid, dimid[0], NULL, &rows), "rows");
  CHK (cw_inq_dim (cwid, dimid[1], NULL, &cols), "columns");
  dvals = rows*cols;

  /* Set default ocean point */
  /* ----------------------- */
  if (guess_ocean) {
    ocean[0] = rows;
    ocean[1] = 1;
  } /* if */
  else {
    if (ocean[0] < 1 || ocean[0] > rows)
      XIT ("ocean row point is out of bounds", "error", -5);
    if (ocean[1] < 1 || ocean[1] > cols)
      XIT ("ocean column point is out of bounds", "error", -5);
  } /* else */

  /* Check external type */
  /* ------------------- */
  if (xtype != CW_FLOAT)
    XIT ("invalid input variable", var_name, -5);

  /* Allocate memory */
  /* --------------- */
  if ((fp = (float *) malloc (sizeof (float)*dvals)) == NULL)
    XIT ("unable to allocate variable memory", "malloc", -6);
  if ((ucp = (unsigned char *) malloc (dvals)) == NULL)
    XIT ("unable to allocate graphics memory", "malloc", -6);
  if ((landmask = (unsigned char *) malloc (dvals)) == NULL)
    XIT ("unable to allocate landmask memory", "malloc", -6);

  /* Read variable and graphics data */
  /* ------------------------------- */
  start[0] = 0;
  start[1] = 0;
  count[0] = rows;
  count[1] = cols;
  CHK (cw_get_vara_float (cwid, varid, start, count, fp), var_name);
  CHK (cw_get_vara_uchar (cwid, gvarid, start, count, ucp), "graphics");
  memcpy (landmask, ucp, dvals);

  /* Flag coastline values */
  /* --------------------- */
  for (i = 0; i < dvals; i++)
    landmask[i] = (landmask[i] & 0x04) + 10;

  /* Create landmask */
  /* --------------- */
  polyfill (ocean[1], ocean[0], 10, OCEAN, LAND, recursive, landmask, 
    rows, cols);  
  for (e = 0; e < extend; e++)
    dilate (landmask, rows, cols, se, 3, OCEAN, LAND, LAND);

  /* Apply land mask to graphics planes and variable data */
  /* ---------------------------------------------------- */
  for (i = 0; i < dvals; i++)
    if (landmask[i] == LAND) {
      fp[i] = CW_BADVAL;
      ucp[i] |= 0x01;
    } /* if */
  CHK (cw_put_vara_float (cwid, varid, start, count, fp), var_name);
  CHK (cw_put_vara_uchar (cwid, gvarid, start, count, ucp), "graphics");

  /* Free memory, close files, and return */
  /* ------------------------------------ */
  free ((void *) fp);
  free ((void *) ucp);
  CHK (cw_close (cwid), ofname);
  return (CW_NOERR);

} /* main */

/**********************************************************************/
