/**********************************************************************/
/*
     FILE: cwfstats.c
  PURPOSE: To print statistics on a CoastWatch format file.
   AUTHOR: Peter Hollemans
     DATE: 28/04/1998
  CHANGES: 25/06/1998, PFH, fixed sum initialization bug

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cwf.h"

/* defined constants */

#define UMESSAGE "\
usage: cwfstats input\n\
where: input            The CoastWatch dataset to read.\n"

#define NARG	1

/* defined macros */

#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 

#define STR(e) cw_strerror ((e))

static int cwerr;
#define CHK(r,v) { \
  if ((cwerr = (r)) != CW_NOERR) \
    XIT (STR (cwerr), (v), cwerr); \
} 

/**********************************************************************/

void err (const char *m, const char *v) {
  fprintf (stderr, "cwfstats: %s: %s\n", v, m);
} /* err */

/**********************************************************************/

void usage (int ecode) {
  fprintf (stderr, UMESSAGE);
  exit (ecode);
} /* usage */

/**********************************************************************/

int main (int argc, char *argv[]) {

char cbuf[32];
int cwid, varid, dimid[2];
size_t rows, cols, start[2], count[2], i, j;
char var_name[CW_MAX_NAME];
float min, max;
double mean, stdev, sum, sdsum;
long dcount, dgood;
float *fp;

  argc--, argv++;				/* parse command line */
  if (argc < NARG)
    usage (-1);

						/* open file */
  CHK (cw_open (argv[0], CW_NOWRITE, &cwid), argv[0]);

						/* get dimensions */
  CHK (cw_inq_dimid (cwid, "rows", dimid), "rows");
  CHK (cw_inq_dimid (cwid, "columns", dimid+1), "columns");
  CHK (cw_inq_dim (cwid, dimid[0], NULL, &rows), "rows");
  CHK (cw_inq_dim (cwid, dimid[1], NULL, &cols), "columns");

  varid = 0;					/* get variable info */
  CHK (cw_inq_var (cwid, varid, var_name, NULL, NULL, NULL, NULL),
    "variable 0");

						/* allocate memory */
  if ((fp = (float *) malloc (sizeof (float)*cols)) == NULL)
    XIT ("unable to allocate data memory", "malloc", -3);

  start[1] = 0;					/* initial data read */
  count[0] = 1;
  count[1] = cols;
  dcount = (long) rows * cols;
  dgood = 0;
  sum = 0;
  for (i = 0; i < rows; i++) {
    start[0] = i;
    CHK (cw_get_vara_float (cwid, varid, start, count, fp), var_name);
    for (j = 0; j < cols; j++) {
      if (fp[j] != CW_BADVAL) {

        dgood++;				/* increment good count */

        if (dgood == 1)				/* get min, max */
          min = max = fp[j];
        else {
          if (fp[j] < min)
            min = fp[j];
          if (fp[j] > max)
            max = fp[j];
        } /* else */

        sum += fp[j];				/* add to sum */

      } /* if */          
    } /* for */
  } /* for */

  if (dgood != 0) {				/* second data read */

    mean = sum/dgood;				/* calc mean */

    if (dgood != 1) {				/* calc stdev */
      sdsum = 0;
      for (i = 0; i < rows; i++) {
        start[0] = i;
        CHK (cw_get_vara_float (cwid, varid, start, count, fp), 
          var_name);
        for (j = 0; j < cols; j++) {
          if (fp[j] != CW_BADVAL) {
            sdsum += pow ((double) fp[j] - mean, 2); 
          } /* if */          
        } /* for */
      } /* for */
      stdev = sqrt (sdsum/((double) dgood - 1));
    } /* if */
    else 
      stdev = 0;

  } /* if */
  else {
    mean = 0;
    stdev = 0;
    min = 0;
    max = 0;
  } /* if */

						/* print stats */
  printf ("Count      Good       Mean       Min        Max        Std Dev\n");
  printf ("%-10ld %-10ld %-#10.7g %-#10.7g %-#10.7g %-#10.7g\n", dcount, 
    dgood, mean, min, max, stdev);

  free ((void *) fp);				/* finish up */
  CHK (cw_close (cwid), argv[0]);
  return (CW_NOERR);

} /* main */

/**********************************************************************/
