/**********************************************************************/
/*
     FILE: cwftoasc.c
  PURPOSE: To convert a CoastWatch format file to lines of longitude, 
           latitude, and ASCII values.
   AUTHOR: Peter Hollemans
     DATE: 21/11/1997
  CHANGES: 26/01/1998, PFH, changed name to cwftoasc for DOS use, replaced
             size_t with unsigned long for DOS TURBO C compiler, added
             check for malloc and IMread_data
           07/05/1998, PFH, rewrote to use new libcwf
           27/08/1998, PFH, corrected mistake in usage note

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "cwf.h"
#include "cwproj.h"

/* defined constants */

#define UMESSAGE "\
usage: cwftoasc [-p fdig] [-d ddig] [-r] input\n\
where: -p fdig          Changes the default precision (number of\n\
                        significant digits) for floating-point values (6).\n\
       -d ddig          Changes the default precision (number of\n\
                        significant digits) for double precision\n\
                        floating-point values (10).\n\
       -r               Reverses the default coordinate order (lon, lat).\n\
       -t               Separates data values with TAB rather than SPACE.\n\
       input            The CoastWatch dataset to read.\n"

#define NARG	1

/* defined macros */

#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 

#define STR(e) cw_strerror ((e))

static int cwerr;
#define CHK(r,v) { \
  if ((cwerr = (r)) != CW_NOERR) \
    XIT (STR (cwerr), (v), cwerr); \
} 

/**********************************************************************/

void err (const char *m, const char *v) {
  fprintf (stderr, "cwftoasc: %s: %s\n", v, m);
} /* err */

/**********************************************************************/

void usage (int ecode) {
  fprintf (stderr, UMESSAGE);
  exit (ecode);
} /* usage */

/**********************************************************************/

int main (int argc, char *argv[]) {

char format[32];
int cwid, varid, dimid[2];
size_t rows, cols, start[2], count[2], i, j;
cw_type var_xtype;
char var_name[CW_MAX_NAME];
float *fp;
double *lat, *lon, *x, *y;
int fdig = 6, ddig = 10;
int sep = ' ';
int rev = 0;

  argc--, argv++;				/* parse command line */
  if (argc < NARG)
    usage (-1);
  while (argv[0][0] == '-') {
    switch (argv[0][1]) {
      case 'p': 
        fdig = atoi (argv[1]); 
        if (fdig < 1)
          XIT ("invalid number of digits", argv[1], -2);
        argc-=2, argv+=2; 
        break;
      case 'd': 
        ddig = atoi (argv[1]); 
        if (ddig < 1)
          XIT ("invalid number of digits", argv[1], -2);
        argc-=2, argv+=2; 
        break;
      case 'r': 
        rev = 1;
        argc--, argv++; 
        break;
      case 't': 
        sep = '\t';
        argc--, argv++; 
        break;
      default: 
        err ("invalid option", argv[0]); 
        usage (-3);
    } /* switch */
    if (argc < NARG)
      usage (-4);
  } /* while */

						/* open files */
  CHK (cw_open (argv[0], CW_NOWRITE, &cwid), argv[0]);

						/* get dimensions */
  CHK (cw_inq_dimid (cwid, "rows", dimid), "rows");
  CHK (cw_inq_dimid (cwid, "columns", dimid+1), "columns");
  CHK (cw_inq_dim (cwid, dimid[0], NULL, &rows), "rows");
  CHK (cw_inq_dim (cwid, dimid[1], NULL, &cols), "columns");

  varid = 0;					/* get variable info */
  CHK (cw_inq_var (cwid, varid, var_name, &var_xtype, NULL, NULL, NULL),
    "variable 0");

  CHK (cw_init_proj (cwid), argv[0]);		/* initialize projection */

						/* allocate memory */
  if ((lat = (double *) malloc (sizeof (double)*cols)) == NULL)
    XIT ("unable to allocate data memory", "malloc", -5);
  if ((lon = (double *) malloc (sizeof (double)*cols)) == NULL)
    XIT ("unable to allocate data memory", "malloc", -6);
  if ((fp = (float *) malloc (sizeof (float)*cols)) == NULL)
    XIT ("unable to allocate data memory", "malloc", -7);

  if (var_xtype == CW_BYTE)			/* create print format */
    sprintf (format, "%%#.%dg%c%%#.%dg%c%%.0f\n", ddig, sep, ddig, sep);
  else if (var_xtype == CW_FLOAT)
    sprintf (format, "%%#.%dg%c%%#.%dg%c%%#.%dg\n", ddig, sep, ddig, sep, fdig);
  else
    XIT ("unknown", "variable external type", -8);

  if (rev) {					/* determine print order */
    x = lat;
    y = lon;
  } /* if */
  else {
    x = lon;
    y = lat;
  } /* else */

  start[1] = 0;					/* print values */
  count[0] = 1;
  count[1] = cols;
  for (i = 0; i < rows; i++) {
    start[0] = i;
    CHK (cw_get_vara_float (cwid, varid, start, count, fp), var_name);
    for (j = 0; j < cols; j++)
      cw_get_ll (j+1, i+1, lat+j, lon+j);
    for (j = 0; j < cols; j++)
      printf (format, x[j], y[j], fp[j]);
  } /* for */

  free ((void *) lat);				/* finish up */
  free ((void *) lon);
  free ((void *) fp);
  CHK (cw_close (cwid), argv[0]);
  return (CW_NOERR);

} /* main */

/**********************************************************************/
