/**********************************************************************/
/*
     FILE: cwftohdf.c
  PURPOSE: To convert a series of IMGMAP-based CoastWatch files of the
           same date and projection to a CoastWatch HDF file.
   AUTHOR: Peter Hollemans
     DATE: 1999/03/01
  CHANGES: 1999/04/06, PFH, shortened some variable attribute names
           1999/06/09, PFH, fixed day/night bug (always indicated day)
           1999/11/07, PFH, reformtted for new C coding style
    NOTES: This routine was rewritten between version 2.2 and 2.3 of
           the library.  This new version writes a completely
           different set of attributes, reformats some variable names,
           and writes data in both compressed and uncompressed format.
           See the file HDFMETA in the library distribution for
           information on the HDF metadata.

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998-1999, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* Include files */
/* ------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "cwf.h"
#include "cwproj.h"
#include "hcwf.h"

/* Usage message constants */
/* ----------------------- */
#define UMESSAGE "\
usage: cwftohcw [-u] [-v] input1 [input2 [input3 [...]]] output\n\
where: -u               Write uncompressed HDF data, default is to use\n\
			compression.\n\
       -v               Print verbose status messages.\n\
       input1 input2 .. The CoastWatch dataset(s) to read.\n\
       output           The CoastWatch HDF dataset to write.\n"
#define NARG		2
#define PROG		"cwftohdf"

/* Sphereoid radius and type constants */
/* ----------------------------------- */
#define R_STD		6370.997
#define R_CW		6371.2
#define NEWSPHEROID	19

/* Variable units string constants */
/* ------------------------------- */
#define IR_UNITS	"celsius"
#define VIS_UNITS	"albedo*100%"
#define TIME_UNITS	"hours"
#define ANG_UNITS	"degrees"

/* Error-checking macros */
/* --------------------- */
#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 
#define STR(e) cw_strerror ((e))

static int cwerr;
#define CCHK(r,v) { \
  if ((cwerr = (r)) != CW_NOERR) \
    XIT (STR (cwerr), (v), cwerr); \
} 
#define HCCHK(r,v) { \
  if ((r) == FAIL) \
    XIT (CWstrerror (), (v), -1) \
} 
#define HCHK(r,v) { \
  if ((r) == FAIL) \
    XIT (HEstring (HEvalue (1)), (v), -1) \
} 

/* Longitude range macro */
/* --------------------- */
#define LONR(a) {	\
  if ((a) >= 180)	\
    (a) -= 360;		\
  else if ((a) < -180)	\
    (a) += 360;		\
}

/* Attribute copy macros */
/* --------------------- */
static char tmp_text_val[50];
#define COPYATT_TEXT(c,h,ain,aout) { \
  CCHK (cw_get_att_text ((c), 0, (ain), tmp_text_val), (ain)) \
  HCHK (SDsetattr ((h), (aout), DFNT_CHAR8, strlen (tmp_text_val), \
    tmp_text_val), (aout)) \
}

static short tmp_short_val;
#define COPYATT_SHORT(c,h,ain,aout) { \
  CCHK (cw_get_att_short ((c), 0, (ain), &tmp_short_val), (ain)) \
  HCHK (SDsetattr ((h), (aout), DFNT_INT16, 1, &tmp_short_val), (aout)) \
}

/**********************************************************************/

void err (			/*** PRINT ERROR MESSAGE ***/
  const char *m,		/* error message string */
  const char *v			/* variable that message applies to */
) {
  
  fprintf (stderr, "%s: %s: %s\n", PROG, v, m);

} /* err */

/**********************************************************************/

void usage (			/*** PRINT USAGE MESSAGE ***/
  int ecode			/* exit status code */
) {

  fprintf (stderr, UMESSAGE);
  exit (ecode);

} /* usage */

/**********************************************************************/

void makedate (			/*** CREATE DATE FROM FILE ***/
  int cwid,			/* CoastWatch file id */
  CWDATE *date			/* date variable (modified) */
) {

short year, day, hour_minute, hour, minute, second, millisecond;

  /* Get date and time from input file */
  /* --------------------------------- */
  CCHK (cw_get_att_short (cwid, 0, "orbit_start_year", &year), 
    "orbit_start_year");
  if (year < 100) {
    if (year >= 70)
      year = 1900 + year;
    else
      year = 2000 + year;
  } /* if */
  CCHK (cw_get_att_short (cwid, 0, "orbit_start_day", &day),
    "orbit_start_day");
  CCHK (cw_get_att_short (cwid, 0, "orbit_start_hour_minute", &hour_minute),
    "orbit_start_hour_minute");
  hour = hour_minute/100;
  minute = hour_minute - hour*100;
  CCHK (cw_get_att_short (cwid, 0, "orbit_start_second", &second),
    "orbit_start_second");
  CCHK (cw_get_att_short (cwid, 0, "orbit_start_millisecond", &millisecond),
    "orbit_start_millisecond");

  /* Create date */
  /* ----------- */
  HCCHK (CWmakedate (year, day, hour, minute, second, millisecond, 
    date), "CWmakedate")

} /* makedate */

/**********************************************************************/

void makeproj (			/*** CREATE EARTH TRANSFORM FROM FILE ***/
  int cwid,			/* CoastWatch file id */
  CWXFM *xfm			/* transform variable (modified) */
) {

proj_info pinfo;
char proj_type[20];
int dimid;
size_t rows, cols;
double clat, clon;
float64 gctp_parm[15];
int32 gctp_sys;
float64 pixel_size;

  /* Initialize projection info */
  /* -------------------------- */
  CCHK (cw_init_proj (cwid), "cw_init_proj")
  CCHK (cw_proj_info (&pinfo), "cw_proj_info")
  CCHK (cw_get_att_text (cwid, 0, "projection_type", proj_type),
    "projection_type")

  /* Get file dimensions */
  /* ------------------- */
  CCHK (cw_inq_dimid (cwid, "rows", &dimid), "rows")
  CCHK (cw_inq_dim (cwid, dimid, NULL, &rows), "rows")
  CCHK (cw_inq_dimid (cwid, "columns", &dimid), "columns")
  CCHK (cw_inq_dim (cwid, dimid, NULL, &cols), "columns")

  /* Get projection center point */
  /* --------------------------- */
  cw_get_ll ((cols+1)/2.0, (rows+1)/2.0, &clat, &clon);

  /* Create GCTP info */
  /* ---------------- */
  switch (pinfo.ptype) {

    /* Polar stereographic projection */
    /* ------------------------------ */
    case POLAR:
    gctp_sys = PS;
    pixel_size = pinfo.res*(R_STD/R_CW)*1000.0;
    LONR (pinfo.plon)
    GCTPPARM (gctp_parm,
      0,0,0,0, 
      GCTPPAK (pinfo.plon), 
      GCTPPAK (pinfo.hem*60.0),
      0,0,0,0,0,0,0,0,0
    )
    break;
   
    /* Mercator projection */
    /* ------------------- */
    case MERCATOR:
    gctp_sys = MERCAT;
    pixel_size = pinfo.res*(R_STD/R_CW)*1000.0;
    GCTPPARM (gctp_parm,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
    break;

    /* Linear lat/lon projection */
    /* ------------------------- */
    case LINEAR:
    gctp_sys = GEO;
    pixel_size = pinfo.res;
    GCTPPARM (gctp_parm,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
    break;

    /* Unknown projection */
    /* ------------------ */
    default:
    XIT ("unsupported projection type", proj_type, -1);
    break;

  } /* case */

  /* Create earth transform */
  /* ---------------------- */
  HCCHK (CWmakeproj (gctp_sys, 0, gctp_parm, NEWSPHEROID, rows, cols, 
    clat, clon, pixel_size, pixel_size, xfm), "CWmakeproj")

} /* makeproj */

/**********************************************************************/

void makeunits (		/*** CREATE UNITS STRING ***/
  int cwid,			/* CoastWatch file id */
  char *var_name,		/* variable name */
  char *units			/* variable units (modified) */
) {

char data_id[20];

  /* Get variable data type */
  /* ---------------------- */
  CCHK (cw_get_att_text (cwid, 0, "data_id", data_id), "data_id")

  /* Create units string */
  /* ------------------- */
  if (strcmp (data_id, "infrared") == 0)
    strcpy (units, IR_UNITS);
  else if (strcmp (data_id, "visible") == 0)
    strcpy (units, VIS_UNITS);
  else if (strcmp (data_id, "ancillary") == 0) {
    if (strcmp (var_name, "scan_time") == 0)
      strcpy (units, TIME_UNITS);
    else
      strcpy (units, ANG_UNITS);
  } /* else if */

} /* makeunits */

/**********************************************************************/

void makename (			/*** CREATE VARIABLE LONG NAME ***/
  char *var_name,		/* variable name */
  char *long_name		/* variable long name (modified) */
) {

int i;
int n_names = 23;
char *var_names[] = {
  "avhrr_ch1",		"avhrr_ch2",		"avhrr_ch3",
  "avhrr_ch4",		"avhrr_ch5",		"mcsst",
  "scan_angle",		"sat_zenith",		"solar_zenith",
  "rel_azimuth",	"scan_time",		"mcsst_split",
  "mcsst_dual",		"mcsst_triple",		"cpsst_split",
  "cpsst_dual",		"cpsst_triple",		"nlsst_split",
  "nlsst_dual",		"nlsst_triple",		"ocean_reflect",
  "turbidity",		"cloud"
};
char *long_names[] = {
  "AVHRR channel 1",
  "AVHRR channel 2",
  "AVHRR channel 3",
  "AVHRR channel 4",         
  "AVHRR channel 5",
  "sea surface temperature",
  "scan angle",
  "satellite zenith angle",   
  "solar zenith angle",
  "relative azimuth angle",
  "scan time",
  "sea surface temperature",
  "sea surface temperature",
  "sea surface temperature",
  "sea surface temperature",
  "sea surface temperature",
  "sea surface temperature",
  "sea surface temperature",
  "sea surface temperature",
  "sea surface temperature",
  "ocean reflectance",
  "turbidity",
  "NESDIS CLAVR cloud mask"
};

  /* Find variable name in the list */
  /* ------------------------------ */
  for (i = 0; i < n_names; i++)
    if (strcmp (var_name, var_names[i]) == 0) {
      strcpy (long_name, long_names[i]);
      break;
    } /* if */

  /* If not found, label unknown */
  /* --------------------------- */
  if (i == n_names)
    strcpy (long_name, "unknown");

} /* makename */

/**********************************************************************/

void makeatts (			/*** CREATE EXTRA VARIABLE ATTRIBUTES ***/
  int cwid, 
  int32 sdsid, 
  char *var_name
) {

char sst_equation[60];

  /* Create visible channel attributes */
  /* --------------------------------- */
  if (!strcmp (var_name, "avhrr_ch1") || !strcmp (var_name, "avhrr_ch2"))
    COPYATT_TEXT (cwid, sdsid, "sun_normalization", "solar_corr")   

  /* Create infrared channel attributes */
  /* ---------------------------------- */
  if (!strcmp (var_name, "avhrr_ch4") || !strcmp (var_name, "avhrr_ch5")) {
    COPYATT_TEXT (cwid, sdsid, "limb_correction", "limb_corr")   
    COPYATT_TEXT (cwid, sdsid, "nonlinearity_correction", 
      "nonlinear_corr")   
  } /* if */

  /* Create SST attributes */
  /* --------------------- */
  if (strstr (var_name, "sst") != NULL) {

    if (strncmp (var_name, "mc", 2) == 0)
      strcpy (sst_equation, "linear multichannel");
    else if (strncmp (var_name, "cp", 2) == 0)
      strcpy (sst_equation, "cross product");
    else if (strncmp (var_name, "nl", 2) == 0)
      strcpy (sst_equation, "nonlinear");

    if (strstr (var_name, "split") != NULL)
      strcat (sst_equation, " split-window");
    else if (strstr (var_name, "dual") != NULL)
      strcat (sst_equation, " dual-window");
    else if (strstr (var_name, "triple") != NULL)
      strcat (sst_equation, " triple-window");

    HCHK (SDsetattr (sdsid, "sst_equation", DFNT_CHAR8, 
      strlen (sst_equation), sst_equation), "sst_equation")

  } /* if */

  /* Create other attributes */
  /* ----------------------- */
  COPYATT_SHORT (cwid, sdsid, "percent_non_zero", "percent_good")

} /* makeatts */

/**********************************************************************/

int main (			/*** MAIN PROGRAM ***/
  int argc,			/* argument count */
  char *argv[]			/* argument values */
) {

int save_argc = argc;
char **save_argv = argv;
char **ifname, *ofname;
int cwid, varid;
CWDATE date, new_date;
CWXFM xfm, new_xfm;
char sat[20], orb[20];
intn day;
int32 sdid, sdsid;
char var_name[CW_MAX_NAME];
cw_type cwvar_xtype;
char units[20];
char long_name[60];
float64 scale;
char save_var_name[CW_MAX_NAME];
float32 *fp;
uint8 *bp;
int32 i, dvals;
size_t start[2], count[2];
int compress = 1, graphics = 0, verbose = 0;
int n_files, k;

  /* Parse command line */
  /* ------------------ */
  argc--, argv++;
  if (argc < NARG)
    usage (-1);
  while (argv[0][0] == '-') {
    switch (argv[0][1]) {
      case 'u': 
        compress = 0;
        argc--, argv++; 
        break;
      case 'v': 
        verbose = 1;
        argc--, argv++; 
        break;
      default: 
        err ("invalid option", argv[0]); 
        usage (-2);
    } /* switch */
    if (argc < NARG)
      usage (-2);
  } /* while */

  n_files = argc-1;
  ifname = argv;
  ofname = argv[argc-1];  


  /* Open first input file */
  /* --------------------- */
  CCHK (cw_open (ifname[0], CW_NOWRITE, &cwid), ifname[0])

  /* Create output file */
  /* ------------------ */
  makedate (cwid, &date);
  makeproj (cwid, &xfm);
  CCHK (cw_get_att_text (cwid, 0, "satellite_id", sat), "satellite_id")
  CCHK (cw_get_att_text (cwid, 0, "orbit_time", orb), "orbit_time")
  day = (strcmp (orb, "day") == 0 ? 1 : 0);
  HCCHK (CWcreate (ofname, sat, "avhrr", xfm, date, day, &sdid), "CWcreate") 

  /* Setup output rows, cols */
  /* ----------------------- */
  dvals = xfm->rows * xfm->cols;
  start[0] = start[1] = 0; 
  count[0] = xfm->rows; count[1] = xfm->cols;

  for (k = 0; k < n_files; k++) {
  
    /* Write verbose message */
    /* --------------------- */
    if (verbose) {
      printf ("%s: converting file [%4d/%4d], %s\n", 
        PROG, k+1, n_files, ifname[k]);
    } /* if */

    /* Get variable info */
    /* ----------------- */
    CCHK (cw_inq_var (cwid, 0, var_name, &cwvar_xtype, NULL, NULL, NULL), 
      "var0")
    makeunits (cwid, var_name, units);
    makename (var_name, long_name);

    switch (cwvar_xtype) {

      /* Transfer byte data */
      /* ------------------ */
      case CW_BYTE:
      HCCHK (CWnewbyte (sdid, var_name, long_name, compress, &sdsid), 
        "CWnewbyte")
      makeatts (cwid, sdsid, var_name);
      if ((bp = malloc (dvals)) == NULL)
        XIT ("unable to allocate data memory", "malloc", -3)
      CCHK (cw_get_vara_uchar (cwid, 0, start, count, bp), var_name);
      HCCHK (CWwritebyte (sdsid, bp), var_name)
      HCHK (SDendaccess (sdsid), var_name)
      free (bp);
      break;

      /* Transfer floating-point data */
      /* ---------------------------- */
      case CW_FLOAT:

      /* Set HDF calibration scaling factor */
      /* ---------------------------------- */
      if (strstr (long_name, "angle") != NULL)
        scale = 0.005;
      else
        scale = 0.01;

      /* Simplify SST variable name */
      /* -------------------------- */
      strcpy (save_var_name, var_name);
      if (strstr (var_name, "sst") != NULL)
        strcpy (var_name, "sst");

      /* Read/write float data */
      /* --------------------- */
      HCCHK (CWnewfloat (sdid, var_name, long_name, units, scale, 
        0.0, compress, &sdsid), "CWnewfloat")
      makeatts (cwid, sdsid, save_var_name);
      if ((fp = malloc (sizeof (float32)*dvals)) == NULL)
        XIT ("unable to allocate data memory", "malloc", -3)
      CCHK (cw_get_vara_float (cwid, 0, start, count, fp), var_name)
      HCCHK (CWwritefloat (sdsid, CW_BADVAL, fp), var_name)
      HCHK (SDendaccess (sdsid), var_name)
      free (fp);
      break;

      /* Unknown variable type */
      /* --------------------- */
      default:
      XIT ("unknown external variable type", var_name, -4);
      break;

    } /* switch */

    /* Transfer graphics data */
    /* ---------------------- */
    if (!graphics) {
      strcpy (var_name, "graphics");
      if (cw_inq_varid (cwid, var_name, &varid) == CW_NOERR) {
        HCCHK (CWnewbyte (sdid, var_name, "graphics planes", compress, 
          &sdsid), "CWnewbyte")
        if ((bp = malloc (dvals)) == NULL)
          XIT ("unable to allocate data memory", "malloc", -5)
        CCHK (cw_get_vara_uchar (cwid, varid, start, count, bp), var_name);
        HCCHK (CWwritebyte (sdsid, bp), var_name)
        HCHK (SDendaccess (sdsid), var_name)
        free (bp);
	graphics = 1;
      } /* if */
    } /* if */

    /* Open next input file */
    /* -------------------- */
    CCHK (cw_close (cwid), ifname[k])
    if (k != n_files-1) {

      /* Check input file date */
      /* --------------------- */
      CCHK (cw_open (ifname[k+1], CW_NOWRITE, &cwid), ifname[k+1])
      makedate (cwid, &new_date);
      if (CWcomparedate (date, new_date) != 0)
        XIT ("input file dates do not match", ifname[k+1], -6)
      free (new_date);
 
      /* Check input file projection */
      /* --------------------------- */
      makeproj (cwid, &new_xfm);
      if (CWsameproj (xfm, new_xfm) != 0)
        XIT ("input file projections do not match", ifname[k+1], -6)
      free (new_xfm);

    } /* if */

  } /* for */

  /* Finish up and return */
  /* ------------------- */
  HCCHK (CWaddhist (sdid, save_argc, save_argv), ofname)
  HCHK (SDend (sdid), ofname)
  return (0);

} /* main */    

/**********************************************************************/
