/**********************************************************************/
/*
     FILE: cwfval.c
  PURPOSE: To print a variable value from a CoastWatch format file.
   AUTHOR: Peter Hollemans
     DATE: 07/06/1998
  CHANGES: 01/09/1998, PFH, added -f option
	   28/02/1999, PFH, simplified round function

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "cwf.h"
#include "cwproj.h"

/* defined constants */

#define UMESSAGE "\
usage: cwfval [-p fdig] [-w] -c x y | -f cfile input\n\
where: -p fdig          Changes the default precision (number of\n\
                        significant digits) for floating-point values (6).\n\
       -w               Specifies that x and y are in real world coordinates\n\
                        (longitude, latitude) rather than image coordinates\n\
                        (column, row).\n\
       -c x,y           Specifies the coordinates of the desired data value.\n\
                        In image coordinates (the default):\n\
                          1 <= x <= columns\n\
                          1 <= y <= rows\n\
                        where rows and columns are the dataset dimensions.\n\
                        In real world coordinates:\n\
                           -360 < x < 360\n\
                            -90 < y <  90\n\
                        and the data value is approximated using the nearest\n\
                        neighbor.\n\
       -f cfile         Specifies a file of x,y coordinate values, one pair\n\
                        per line, separated by a space.\n\
       input            The CoastWatch dataset to read.\n"

#define NARG	1

/* defined macros */

#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 

#define STR(e) cw_strerror ((e))

static int cwerr;
#define CHK(r,v) { \
  if ((cwerr = (r)) != CW_NOERR) \
    XIT (STR (cwerr), (v), cwerr); \
} 

/**********************************************************************/

void err (const char *m, const char *v) {
  fprintf (stderr, "cwfval: %s: %s\n", v, m);
} /* err */

/**********************************************************************/

void usage (int ecode) {
  fprintf (stderr, UMESSAGE);
  exit (ecode);
} /* usage */

/**********************************************************************/

static int round (float f) {

/* Rounds a float to the nearest integer. */

  return (f > 0 ? floor (f+0.5) : ceil (f-0.5));

} /* round */

/**********************************************************************/

void set_start (size_t *start, double x, double y, int world) {

/* Sets up the start vector for reading. */

double i, j;

  if (world) {					/* world coordinates */
    cw_get_ij (&i, &j, y, x);
    start[0] = round (j) - 1;
    start[1] = round (i) - 1;
  } /* if */

  else {					/* image coordinates */
    start[0] = round (y) - 1;
    start[1] = round (x) - 1;
  } /* else */  

} /* set_start */

/**********************************************************************/

int main (int argc, char *argv[]) {

char format[32];
float fval;
int cwid, varid;
char var_name[CW_MAX_NAME];
cw_type var_xtype;
int world = 0;
size_t start[2], count[2];
char *fname;
double x, y;
int fdig = 6;
FILE *cfp = NULL;
char *cfname = NULL;

  argc--, argv++;				/* parse command line */
  if (argc < NARG)
    usage (-1);
  while (argv[0][0] == '-' && argc > NARG) {
    switch (argv[0][1]) {

      case 'p': 
        fdig = atoi (argv[1]); 
        if (fdig < 1)
          XIT ("invalid number of digits", argv[1], -2);
        argc-=2, argv+=2; 
        break;

      case 'w':
        world = 1;
        argc--, argv++; 
        break;

      case 'c':
        sscanf (argv[1], "%lf,%lf", &x, &y);
        argc-=2, argv+=2; 
        break;

      case 'f':
        cfname = argv[1];
        argc-=2, argv+=2; 
        break;

      default: 
        err ("invalid option", argv[0]); 
        usage (-3);

    } /* switch */
    if (argc < NARG)
      usage (-4);
  } /* while */

  fname = argv[0];
						/* open files */
  CHK (cw_open (fname, CW_NOWRITE, &cwid), fname);
  if (cfname != NULL && (cfp = fopen (cfname, "r")) == NULL)
    XIT ("can't open coordinates file", cfname, -5);

  varid = 0;					/* get variable info */
  CHK (cw_inq_var (cwid, varid, var_name, &var_xtype, NULL, NULL, NULL), 
    "variable 0");

  if (world)					/* initialize projection */
    CHK (cw_init_proj (cwid), fname);

  if (var_xtype == CW_BYTE)			/* create print format */
    sprintf (format, "%%.0f\n");
  else if (var_xtype == CW_FLOAT)
    sprintf (format, "%%#.%dg\n", fdig);
  else
    XIT ("unknown", "variable external type", -6);

  count[0] = count[1] = 1;			/* set count vector */

  if (cfname != NULL) {				/* use file (x, y) */
    fscanf (cfp, "%lf %lf", &x, &y);
    while (!feof (cfp)) {
      set_start (start, x, y, world);
      CHK (cw_get_vara_float (cwid, varid, start, count, &fval), var_name);
      printf (format, fval);
      fscanf (cfp, "%lf %lf", &x, &y);
    } /* while */
  } /* if */

  else {					/* use command line (x, y) */
    set_start (start, x, y, world);
    CHK (cw_get_vara_float (cwid, varid, start, count, &fval), var_name);
    printf (format, fval);
  } /* else */

  CHK (cw_close (cwid), fname);			/* finish up */
  if (cfname != NULL)
    fclose (cfp);
  return (CW_NOERR);

} /* main */

/**********************************************************************/
