/**********************************************************************/
/*
     FILE: hcwf.c
  PURPOSE: To define the CoastWatch HDF library routines.
   AUTHOR: Peter Hollemans
     DATE: 1999/02/23
  CHANGES: 1999/06/15, PFH, added CWclone, SDattr
           1999/11/07, PFH, fixed CWmakedate return

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998-1999, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>

#include "hcwf.h"

/* defined constants */

#define HCW_ETYPE_HDF 0					/* error types */
#define HCW_ETYPE_CW  1

#define HCW_ERR_NUM   6					/* error codes */

#define HCW_ERR_NONE  0
#define HCW_ERR_GCTP  1
#define HCW_ERR_UHDF  2
#define HCW_ERR_MEM   3
#define HCW_ERR_CTYPE 4
#define HCW_ERR_BTYPE 5

#define PI            3.141592653589793238		/* PI */

#define ORIGIN	      "USDOC/NOAA/NESDIS CoastWatch"

/* defined macros */
							/* error checks */
#define HCHK(r) { \
  if ((r) == FAIL) { \
    etype = HCW_ETYPE_HDF; \
    ecode = HEvalue (1); \
    if (ecode == DFE_NONE) { \
      etype = HCW_ETYPE_CW; \
      ecode = HCW_ERR_UHDF; \
    } else { \
      etype = HCW_ETYPE_HDF; \
    } \
    return (FAIL); \
  } \
}
#define CWERR(r) { \
  ecode = (r); \
  etype = HCW_ETYPE_CW; \
  return (FAIL); \
}
#define CCHK(r) { \
  if ((ecode = (r)) == FAIL) { \
    etype = HCW_ETYPE_CW; \
    return (FAIL); \
  } \
}

							/* att read/write */
#define WRITEATT(o,n,t,c,p) { \
  HCHK (SDsetattr ((o), (n), (t), (c), (VOIDP) (p))) \
}

#define DTOR(a)		((a)*PI/180.0)			/* rad conversions */
#define RTOD(a)		((a)*180.0/PI)

							/* round function */
#define RND(f)		((f) > 0 ? floor ((f)+0.5) : ceil ((f)-0.5))

/* global variables */

static long (*g_for_trans[MAXPROJ+1])();		/* Earth xfm info */
static long (*g_inv_trans[MAXPROJ+1])();
static long gctp_err;
static float64 g_et_affine[6], g_et_affine_inv[6];
static int32 g_gctp_sys;

static int etype = HCW_ETYPE_CW;			/* error status */
static int32 ecode = 0;

static char *hcw_error_table[] = {			/* error table */
  "no error",
  "error in cartographic transformation",
  "unknown HDF error",
  "memory allocation failed",
  "unsupported HDF type for calibrated float data",
  "unsupported HDF type for byte data"
};

static char *proj_names[] = {
  "geographic",
  "universal transverse mercator",
  "state plane coordinates",
  "albers conical equal area",
  "lambert conformal conic",
  "mercator",
  "polar stereographic",
  "polyconic",
  "equidistant conic",
  "transverse mercator",
  "stereographic",
  "lambert azimuthal equal area",
  "azimuthal equidistant",
  "gnomonic",
  "orthographic",
  "general vertical near-side perspective",
  "sinusiodal",
  "equirectangular",
  "miller cylindrical",
  "van der grinten",
  "(hotine) oblique mercator",
  "robinson",
  "space oblique mercator",
  "alaska conformal",
  "interrupted goode homolosine ",
  "mollweide",
  "interrupted mollweide",
  "hammer",
  "wagner iv",
  "wagner vii",
  "oblated equal area"
};

/**********************************************************************/

static intn SDattr (		/*** GET HDF SD ATTRIBUTE ***/
  int32 objid,			/* HDF object ID */
  char *att,			/* attribute name */
  VOIDP att_val			/* attribute value (returned) */
) {

int32 att_index;
char att_name[MAX_NC_NAME];
int32 data_type, n_values;
char *string;

  /* Check attribute exists */
  /* ---------------------- */
  if ((att_index = SDfindattr (objid, att)) == FAIL)
    return (FAIL);

  /* Get attribute value */
  /* ------------------- */
  HCHK (SDattrinfo (objid, att_index, att_name, &data_type, &n_values))
  HCHK (SDreadattr (objid, att_index, att_val))
  if (data_type == DFNT_CHAR8) {
    string = att_val;
    string[n_values] = '\0';    
  } /* if */

  return (SUCCEED);

} /* SDattr */

/**********************************************************************/

intn CWcreate (			/*** CREATE A NEW COASTWATCH HDF FILE ***/
  char *path,			/* file path and name */
  char *sat,			/* satellite name, eg: noaa-14, noaa-15 */
  char *sensor,			/* sensor name, eg: avhrr, seawifs */
  CWXFM xfm,                    /* earth transform obtained from 
                                   CWgetproj or CWmakeproj */
  CWDATE date,			/* date obtained from 
				   CWgetdate or CWmakedate */
  intn day,			/* true if the data is from a daytime 
				   satellite pass, false otherwise */
  int32 *sdidp			/* pointer to HDF SD interface id (returned) */
) {

char *projection;

  HCHK (*sdidp = SDstart (path, DFACC_CREATE))		/* create file */

							/* set instruments */
  WRITEATT (*sdidp, "satellite", DFNT_CHAR8, strlen (sat), sat)
  WRITEATT (*sdidp, "sensor", DFNT_CHAR8, strlen (sensor), sensor)

							/* set date/time */
  WRITEATT (*sdidp, "pass_date", DFNT_INT32, 1, &(date->pass_date))
  WRITEATT (*sdidp, "start_time", DFNT_FLOAT64, 1, &(date->start_time))
  if (day)
    WRITEATT (*sdidp, "pass_type", DFNT_CHAR8, 3, "day")
  else
    WRITEATT (*sdidp, "pass_type", DFNT_CHAR8, 5, "night")

							/* set projection */
  projection = proj_names[xfm->gctp_sys];
  WRITEATT (*sdidp, "projection", DFNT_CHAR8, strlen (projection), projection)
  WRITEATT (*sdidp, "gctp_sys", DFNT_INT32, 1, &(xfm->gctp_sys))
  WRITEATT (*sdidp, "gctp_zone", DFNT_INT32, 1, &(xfm->gctp_zone))
  WRITEATT (*sdidp, "gctp_parm", DFNT_FLOAT64, 15, xfm->gctp_parm)
  WRITEATT (*sdidp, "gctp_datum", DFNT_INT32, 1, &(xfm->gctp_datum))
  WRITEATT (*sdidp, "et_affine", DFNT_FLOAT64, 6, xfm->et_affine)
  WRITEATT (*sdidp, "rows", DFNT_INT32, 1, &(xfm->rows))
  WRITEATT (*sdidp, "cols", DFNT_INT32, 1, &(xfm->cols))

							/* set origin */  
  WRITEATT (*sdidp, "origin", DFNT_CHAR8, strlen (ORIGIN), ORIGIN)

  return (SUCCEED);					/* finish up */

} /* CWcreate */

/**********************************************************************/

intn CWclone (			/*** CLONE A COASTWATCH HDF FILE ***/
  char *path,			/* file path and name of existing file */
  char *clone,			/* file path and name of clone to create */
  int32 *sdidp			/* pointer to HDF SD interface id (returned) */
) {

int32 sdid_orig;
char sat[20];
char sensor[20];
CWXFM xfm;
CWDATE date;
intn day;
char pass_type[20];
char hist[2048];
  
  HCHK (sdid_orig = SDstart (path, DFACC_READ))		/* open original */

  HCHK (SDattr (sdid_orig, "satellite", sat))		/* read atts */
  HCHK (SDattr (sdid_orig, "sensor", sensor))
  CCHK (CWgetproj (sdid_orig, &xfm))
  CCHK (CWgetdate (sdid_orig, &date))
  HCHK (SDattr (sdid_orig, "pass_type", pass_type))
  day = (strcmp (pass_type, "day") == 0 ? 1 : 0);  
  HCHK (SDattr (sdid_orig, "history", hist))

							/* create file */
  CCHK (CWcreate (clone, sat, sensor, xfm, date, day, sdidp))
  WRITEATT (*sdidp, "history", DFNT_CHAR8, strlen (hist), hist)

  free (xfm);						/* finish up */
  free (date);
  HCHK (SDend (sdid_orig))
  return (SUCCEED);

} /* CWclone */

/**********************************************************************/

const char *CWstrerror (	/*** RETURN A DESCRIPTIVE ERROR STRING ***/
  void
) {

static char unknown[] = "unknown error";

  if (etype == HCW_ETYPE_HDF)				/* return HDF string */
    return (HEstring (ecode));

  else {						/* return CW string */
    if (ecode >= 0 && ecode < HCW_ERR_NUM)
      return (hcw_error_table[ecode]);
    else
      return (unknown);
  } /* else */

} /* CWstrerror */

/**********************************************************************/

intn CWgetproj (		/*** GET AN EARTH TRANSFORM FROM A FILE ***/
  int32 sdid,			/* HDF SD interface id */
  CWXFM *xfm			/* earth transform (returned) */
) {

  if ((*xfm = malloc (sizeof (cwxfm_struct))) == NULL)	/* allocate memory */
    CWERR (HCW_ERR_MEM)

  HCHK (SDattr (sdid, "gctp_sys", &((*xfm)->gctp_sys)))	/* read xfm atts */
  HCHK (SDattr (sdid, "gctp_zone", &((*xfm)->gctp_zone)))
  HCHK (SDattr (sdid, "gctp_parm", (*xfm)->gctp_parm))
  HCHK (SDattr (sdid, "gctp_datum", &((*xfm)->gctp_datum)))
  HCHK (SDattr (sdid, "et_affine", (*xfm)->et_affine))
  HCHK (SDattr (sdid, "rows", &((*xfm)->rows)))
  HCHK (SDattr (sdid, "cols", &((*xfm)->cols)))

  return (SUCCEED);					/* finish up */

} /* CWgetproj */

/**********************************************************************/

intn CWmakeproj (		/*** CREATE AN EARTH TRANSFORM ***/
  int32 gctp_sys,		/* defines the output projection system */
  int32 gctp_zone,		/* zone for UTM projection system */
  float64 *gctp_parm,		/* array of projection parameters */
  int32 gctp_datum,		/* spheriod code */
  int32 rows, int32 cols,	/* image rows and columns */
  float64 clat, float64 clon,	/* image center point (lat, lon) in degrees,
                                   positive for north and east, negative for
                                   south and west */
  float64 width,		/* image pixel height at the projection 
				   reference point (meters) */
  float64 height,		/* image pixel width at the projection
				   reference point (meters) */
  CWXFM *xfm			/* earth transform (returned) */
) {

int i;
float64 cx, cy;
long (*for_trans[MAXPROJ+1])();

  if ((*xfm = malloc (sizeof (cwxfm_struct))) == NULL)	/* allocate memory */
    CWERR (HCW_ERR_MEM)

  (*xfm)->gctp_sys = gctp_sys;				/* copy GCTP info */
  (*xfm)->gctp_zone = gctp_zone;
  for (i = 0; i < 15; i++)
    (*xfm)->gctp_parm[i] = gctp_parm[i];  
  (*xfm)->gctp_datum = gctp_datum;

  if (gctp_sys == GEO) {				/* find center (x,y) */
    cx = clon;
    cy = clat;
  } /* if */
  else {
    for_init (gctp_sys, gctp_zone, gctp_parm,
      gctp_datum, "", "", &gctp_err, for_trans);
    if (for_trans[gctp_sys] == NULL)
      CWERR (HCW_ERR_GCTP)
    if (for_trans[gctp_sys](DTOR (clon), DTOR (clat), &cx, &cy) != 0)
      CWERR (HCW_ERR_GCTP)
  } /* else */

  (*xfm)->et_affine[0] = width;				/* fill in affine xfm */
  (*xfm)->et_affine[1] = 0.0;
  (*xfm)->et_affine[2] = 0.0;
  (*xfm)->et_affine[3] = -height;
  (*xfm)->et_affine[4] = cx - width*(cols+1)/2.0;
  (*xfm)->et_affine[5] = cy + height*(rows+1)/2.0;

  (*xfm)->rows = rows;					/* copy other info */
  (*xfm)->cols = cols;  

  return (SUCCEED);					/* finish up */

} /* CWmakeproj */

/**********************************************************************/

intn CWuseproj (		/*** USE EARTH TRANFORM IN CALCULATIONS ***/
  CWXFM xfm			/* earth transform to use in subsequent 
				   calls to CWijtoll and CWlltoij */
) {

int i;
float64 det;

  if (xfm->gctp_sys != GEO) {				/* initialize GCTP */

    for_init (xfm->gctp_sys, xfm->gctp_zone, xfm->gctp_parm,
      xfm->gctp_datum, "", "", &gctp_err, g_for_trans);
    if (g_for_trans[xfm->gctp_sys] == NULL)
      CWERR (HCW_ERR_GCTP)

    inv_init (xfm->gctp_sys, xfm->gctp_zone, xfm->gctp_parm,
      xfm->gctp_datum, "", "", &gctp_err, g_inv_trans);
    if (g_inv_trans[xfm->gctp_sys] == NULL)
      CWERR (HCW_ERR_GCTP)

  } /* if */
  g_gctp_sys = xfm->gctp_sys;

  for (i = 0; i < 6; i++)				/* copy et_affine */
    g_et_affine[i] = xfm->et_affine[i];

							/* invert et_affine */
  det = g_et_affine[0]*g_et_affine[3] - g_et_affine[1]*g_et_affine[2];
  g_et_affine_inv[0] = g_et_affine[3] / det;
  g_et_affine_inv[1] = -g_et_affine[1] / det;
  g_et_affine_inv[2] = -g_et_affine[2] / det;
  g_et_affine_inv[3] = g_et_affine[0] / det;
  g_et_affine_inv[4] = -(g_et_affine_inv[0]*g_et_affine[4] + 
                         g_et_affine_inv[1]*g_et_affine[5]);
  g_et_affine_inv[5] = -(g_et_affine_inv[2]*g_et_affine[4] + 
                         g_et_affine_inv[3]*g_et_affine[5]);

  return (SUCCEED);					/* finish up */

} /* CWuseproj */

/**********************************************************************/

intn CWsameproj (		/*** COMPARE EARTH TRANSFORMS */
  CWXFM xfma,			/* earth transform A */
  CWXFM xfmb			/* earth transform B */
				/* NOTE: Returns SUCCEED (0) if transforms
				   match, FAIL (-1) if not. */
) {

int i;

  if (xfma->gctp_sys != xfmb->gctp_sys)			/* check each field */
    return (FAIL);
  if (xfma->gctp_zone != xfmb->gctp_zone)
    return (FAIL);
  for (i = 0; i < 15; i++)
    if (xfma->gctp_parm[i] != xfmb->gctp_parm[i])
      return (FAIL);
  if (xfma->gctp_datum != xfmb->gctp_datum)
    return (FAIL);
  for (i = 0; i < 6; i++)
    if (xfma->et_affine[i] != xfmb->et_affine[i])
      return (FAIL);
  if (xfma->rows != xfmb->rows)
    return (FAIL);
  if (xfma->cols != xfmb->cols)
    return (FAIL);

  return (SUCCEED);					/* finish up */

} /* CWsameproj */

/**********************************************************************/

intn CWijtoll (			/*** CONVERT FROM IMAGE TO GEOGRAPHIC ***/
  float64 i, float64 j,		/* image coordinates, i = column, j = row,
				   values are 1-relative */
  float64 *lat, float64 *lon	/* geographic coordinates, lat = latitude,
				   lon = longitude (returned) */
) {

double x, y;

							/* apply affine */
  x = g_et_affine[0]*i + g_et_affine[1]*j + g_et_affine[4];
  y = g_et_affine[2]*i + g_et_affine[3]*j + g_et_affine[5];

  if (g_gctp_sys == GEO) {				/* call GCTP */
    *lat = y;
    *lon = x;
  } /* if */
  else {
    if (g_inv_trans[g_gctp_sys](x, y, lon, lat) != 0)
      CWERR (HCW_ERR_GCTP)
    *lat = RTOD (*lat);
    *lon = RTOD (*lon);
  } /* else */

  return (SUCCEED);					/* finish up */

} /* CWijtoll */

/**********************************************************************/

intn CWlltoij (			/*** CONVERT FROM GEOGRAPHIC TO IMAGE ***/
  float64 *i, float64 *j,	/* image coordinates, i = column, j = row,
				   values are 1-relative (returned) */
  float64 lat, float64 lon	/* geographic coordinates, lat = latitude,
				   lon = longitude */
) {

double x, y;

  if (g_gctp_sys == GEO) {				/* call GCTP */
    y = lat;
    x = lon;
  } /* if */
  else {
    if (g_for_trans[g_gctp_sys](DTOR (lon), DTOR (lat), &x, &y) != 0)
      CWERR (HCW_ERR_GCTP)
  } /* else */
							/* apply affine */
  *i = g_et_affine_inv[0]*x + g_et_affine_inv[1]*y + g_et_affine_inv[4];
  *j = g_et_affine_inv[2]*x + g_et_affine_inv[3]*y + g_et_affine_inv[5];

  return (SUCCEED);					/* finish up */

} /* CWlltoij */

/**********************************************************************/

intn CWgetdate (		/*** GET A DATE FROM A FILE ***/
  int32 sdid,			/* HDF SD interface id */
  CWDATE *date			/* date structure (returned) */
) {

							/* allocate memory */
  if ((*date = malloc (sizeof (cwdate_struct))) == NULL)
    CWERR (HCW_ERR_MEM)

							/* read data atts */
  HCHK (SDattr (sdid, "pass_date", &((*date)->pass_date)))
  HCHK (SDattr (sdid, "start_time", &((*date)->start_time)))

  return (SUCCEED);					/* finish up */

} /* CWgetdate */

/**********************************************************************/

intn CWmakedate (		/*** CREATE A DATE ***/
  intn year,			/* full year, eg: 1999 */
  intn jday,			/* julian day (1-366) */
  intn hour,			/* hour (0-23) */
  intn minute,			/* minute (0-59) */
  intn second,			/* second (0-59) */
  intn millisecond,		/* millisecond (0-999) */
  CWDATE *date			/* date structure (returned) */
) {

intn ydiff;
							/* allocate memory */
  if ((*date = malloc (sizeof (cwdate_struct))) == NULL)
    CWERR (HCW_ERR_MEM)

  ydiff = year - 1970;					/* create date */
  (*date)->pass_date = (int32) ydiff*365 + (ydiff+1)/4 + (jday-1);
  (*date)->start_time = hour*3600.0 + minute*60.0 + second + 
    millisecond/1000.0;

  return (SUCCEED);					/* finish up */

} /* CWmakedate */

/**********************************************************************/

void CWdatestr (		/*** CREATE A DESCRIPTIVE DATE STRING ***/
  char *date_str,		/* date string char array (returned) */
  size_t buf,			/* date string array size */
  char *date_format,		/* date format (see strftime(3C)) */
  intn local,			/* true for conversion to local time, 
				   false for UTC time */
  CWDATE date			/* date structure */
) {

time_t utc;
struct tm *stmp;

  utc = (date->pass_date)*86400.0 + (date->start_time);
  if (local)
    stmp = localtime (&utc);
  else
    stmp = gmtime (&utc);
  strftime (date_str, buf, date_format, stmp);

} /* CWdatestr */

/**********************************************************************/

intn CWcomparedate (		/*** COMPARE DATES ***/
  CWDATE datea,			/* date A */
  CWDATE dateb			/* date B */
				/* NOTE: Returns -1 if A < B, 0 if A = B,
				   or 1 if A > B. */
) {

time_t utca, utcb;

  utca = (datea->pass_date)*86400.0 + (datea->start_time);
  utcb = (dateb->pass_date)*86400.0 + (dateb->start_time);
  if (utca < utcb)
    return (-1);
  else if (utca > utcb)
    return (1);
  else
    return (0);

} /* CWfirstdate */

/**********************************************************************/

intn CWaddhist (		/*** ADD TO GLOBAL HISTORY ATTRIBUTE ***/
  int32 sdid,			/* HDF SD interface id */
  int argc, char *argv[]	/* C program command line parameters */
) {

int32 aindex;
int32 nvals;
char *hist_buf;
char *prog;
int hist_size, i;
char attr_name[MAX_NC_NAME];
int32 data_type;

  prog = argv[0] + strlen (argv[0]);			/* find prog name */
  while (prog != argv[0] && *(prog-1) != '/' && *(prog-1) != '\\')
    prog--;

  hist_size = strlen (prog)+1;				/* find hist size */
  for (i = 1; i < argc; i++)
    hist_size += strlen (argv[i])+1; 
  hist_size--;

  aindex = SDfindattr (sdid, "history");		/* check history */
  if (aindex == FAIL) {
    hist_buf = (char *) malloc (hist_size+1);
    hist_buf[0] = '\0';    
  } /* if */

  else {						/* read history */
    HCHK (SDattrinfo (sdid, aindex, attr_name, &data_type, &nvals))
    hist_buf = (char *) malloc (nvals+1 + hist_size+1);
    HCHK (SDreadattr (sdid, aindex, (VOIDP) hist_buf))
    hist_buf[nvals] = '\n';
    hist_buf[nvals+1] = '\0';
  } /* else */

  strcat (hist_buf, prog);				/* append history */
  strcat (hist_buf, " ");
  for (i = 1; i < argc; i++) {
    strcat (hist_buf, argv[i]);
    strcat (hist_buf, " ");
  } /* for */
  hist_buf[strlen (hist_buf)-1] = '\0';
  WRITEATT (sdid, "history", DFNT_CHAR8, strlen (hist_buf), hist_buf)

  free ((void *) hist_buf);				/* finish up */
  return (SUCCEED);

} /* CWaddhist */

/**********************************************************************/

intn CWnewfloat (		/*** DEFINE NEW SATELLITE FLOAT VARIABLE ***/
  int32 sdid,			/* HDF SD interface id */
  char *sds_name,		/* HDF scientific dataset variable name */
  char *long_name,		/* descriptive variable name */
  char *units,			/* variable units */
  float64 scale,		/* calibration scale */
  float64 offset,		/* calibration offset */
  intn compress,		/* true to turn on GZIP compression, false
				   for no compression */
  int32 *sdsidp			/* pointer to HDF SD dataset id (returned) */
) {

int32 dim_sizes[2];
int32 gctp_sys;
comp_info c_info;
int32 dim_id;
int16 missing = -32768;

  HCHK (SDattr (sdid, "rows", dim_sizes))		/* create variable */
  HCHK (SDattr (sdid, "cols", dim_sizes+1))
  HCHK (*sdsidp = SDcreate (sdid, sds_name, DFNT_INT16, 2, dim_sizes))

  HCHK (dim_id = SDgetdimid (*sdsidp, 0))		/* set dim names */
  HCHK (SDsetdimname (dim_id, "rows"));
  HCHK (dim_id = SDgetdimid (*sdsidp, 1))
  HCHK (SDsetdimname (dim_id, "cols"));

  HCHK (SDattr (sdid, "gctp_sys", &gctp_sys))		/* set predefined */
  HCHK (SDsetdatastrs (*sdsidp, long_name, units, "F7.2", proj_names[gctp_sys])) 
  HCHK (SDsetfillvalue (*sdsidp, &missing))
  HCHK (SDsetcal (*sdsidp, scale, 0.0, offset, 0.0, DFNT_FLOAT32))

  if (compress) {					/* set compression */
    c_info.deflate.level = 6;  
    HCHK (SDsetcompress (*sdsidp, COMP_CODE_DEFLATE, &c_info))
  } /* if */

							/* set other atts */
  WRITEATT (*sdsidp, "C_format", DFNT_CHAR8, 5, "%7.2f")
  WRITEATT (*sdsidp, "missing_value", DFNT_INT16, 1, &missing)

  return (SUCCEED);					/* finish up */

} /* CWnewfloat */

/**********************************************************************/

intn CWreadfloat (		/*** READ SATELLITE FLOAT DATA ***/
  int32 sdsid,			/* HDF SD dataset id */
  float32 badval,		/* missing value for uncalibrated data */
  float32 **fp			/* pointer to uncalibrated data array (memory 
				   will be allocated for the array) */
) {

char sds_name[MAX_NC_NAME];
int32 dim_sizes[MAX_VAR_DIMS];
int32 rank, cal_data_type, n_attrs;
int32 i, dvals;
int16 *cal_data;
float32 *uncal_data;
int16 missing;
float64 cal, cal_error, offset, offset_err;
int32 data_type;
int32 start[MAX_VAR_DIMS];

							/* allocate memory */ 
  HCHK (SDgetinfo (sdsid, sds_name, &rank, dim_sizes, 
    &cal_data_type, &n_attrs))
  if (cal_data_type != DFNT_INT16)
    CWERR (HCW_ERR_CTYPE)
  for (i = 0, dvals = 1; i < rank; i++)
    dvals = dvals*dim_sizes[i];    
  if ((cal_data = malloc (dvals*sizeof (int16))) == NULL)
    CWERR (HCW_ERR_MEM)
  if ((uncal_data = malloc (dvals*sizeof (float32))) == NULL)
    CWERR (HCW_ERR_MEM)

							/* get attributes */
  HCHK (SDgetcal (sdsid, &cal, &cal_error, &offset, &offset_err, &data_type))
  HCHK (SDattr (sdsid, "missing_value", &missing))

  for (i = 0; i < rank; i++)				/* read data */
    start[i] = 0;
  HCHK (SDreaddata (sdsid, start, NULL, dim_sizes, cal_data))

  for (i = 0; i < dvals; i++) {				/* uncalibrate data */
    if (cal_data[i] == missing)
      uncal_data[i] = badval;
    else 
      uncal_data[i] = cal*(cal_data[i] - offset);
  } /* for */

  *fp = uncal_data;					/* finish up */
  free (cal_data);
  return (SUCCEED);

} /* CWreadfloat */

/**********************************************************************/

intn CWwritefloat (		/*** WRITE SATELLITE FLOAT DATA ***/
  int32 sdsid,			/* HDF SD dataset id */
  float32 badval,		/* missing value for uncalibrated data */
  float32 *fp			/* uncalibrated data array */
) {

char sds_name[MAX_NC_NAME];
int32 dim_sizes[MAX_VAR_DIMS];
int32 rank, cal_data_type, n_attrs;
int32 i, dvals;
int16 *cal_data;
float32 *uncal_data = fp;
int16 missing;
float64 cal, cal_error, offset, offset_err;
int32 data_type;
int32 start[MAX_VAR_DIMS];
float64 test_val;

							/* allocate memory */ 
  HCHK (SDgetinfo (sdsid, sds_name, &rank, dim_sizes, 
    &cal_data_type, &n_attrs))
  if (cal_data_type != DFNT_INT16)
    CWERR (HCW_ERR_CTYPE)
  for (i = 0, dvals = 1; i < rank; i++)
    dvals = dvals*dim_sizes[i];    
  if ((cal_data = malloc (dvals*sizeof (int16))) == NULL)
    CWERR (HCW_ERR_MEM)

							/* get attributes */
  HCHK (SDgetcal (sdsid, &cal, &cal_error, &offset, &offset_err, &data_type))
  HCHK (SDattr (sdsid, "missing_value", &missing))

  for (i = 0; i < dvals; i++) {				/* calibrate data */
    if (uncal_data[i] == badval)
      cal_data[i] = missing;
    else {
      test_val = RND (uncal_data[i] / cal + offset);
      if (test_val > 32767 || test_val < -32767)
        cal_data[i] = missing;
      else
        cal_data[i] = test_val;
    } /* else */
  } /* for */

  for (i = 0; i < rank; i++)				/* write data */
    start[i] = 0;
  HCHK (SDwritedata (sdsid, start, NULL, dim_sizes, cal_data))

  free (cal_data);					/* finish up */
  return (SUCCEED);

} /* CWwritefloat */

/**********************************************************************/

intn CWnewbyte (		/*** DEFINE NEW SATELLITE BYTE VARIABLE ***/
  int32 sdid,			/* HDF SD interface id */
  char *sds_name,		/* HDF scientific dataset variable name */
  char *long_name,		/* descriptive variable name */
  intn compress,		/* true to turn on GZIP compression, false
				   for no compression */
  int32 *sdsidp			/* pointer to HDF SD dataset id (returned) */
) {

int32 dim_sizes[2];
int32 gctp_sys;
comp_info c_info;
int32 dim_id;

  HCHK (SDattr (sdid, "rows", dim_sizes))		/* create variable */
  HCHK (SDattr (sdid, "cols", dim_sizes+1))
  HCHK (*sdsidp = SDcreate (sdid, sds_name, DFNT_UINT8, 2, dim_sizes))

  HCHK (dim_id = SDgetdimid (*sdsidp, 0))		/* set dim names */
  HCHK (SDsetdimname (dim_id, "rows"));
  HCHK (dim_id = SDgetdimid (*sdsidp, 1))
  HCHK (SDsetdimname (dim_id, "cols"));

  HCHK (SDattr (sdid, "gctp_sys", &gctp_sys))		/* set predefined */
  HCHK (SDsetdatastrs (*sdsidp, long_name, "", "I3", proj_names[gctp_sys])) 

  if (compress) {					/* set compression */
    c_info.deflate.level = 6;  
    HCHK (SDsetcompress (*sdsidp, COMP_CODE_DEFLATE, &c_info))
  } /* if */

  WRITEATT (*sdsidp, "C_format", DFNT_CHAR8, 3, "%3u")	/* set other atts */

  return (SUCCEED);					/* finish up */

} /* CWnewbyte */

/**********************************************************************/

intn CWreadbyte (		/*** READ SATELLITE BYTE DATA ***/
  int32 sdsid,			/* HDF SD dataset id */
  uint8 **bp			/* pointer to byte array (memory 
				   will be allocated for the array) */
) {

char sds_name[MAX_NC_NAME];
int32 dim_sizes[MAX_VAR_DIMS];
int32 rank, data_type, n_attrs;
int32 i, dvals;
uint8 *data;
int32 start[MAX_VAR_DIMS];

							/* allocate memory */ 
  HCHK (SDgetinfo (sdsid, sds_name, &rank, dim_sizes, &data_type, &n_attrs))
  if (data_type != DFNT_UINT8)
    CWERR (HCW_ERR_BTYPE)
  for (i = 0, dvals = 1; i < rank; i++)
    dvals = dvals*dim_sizes[i];    
  if ((data = malloc (dvals*sizeof (uint8))) == NULL)
    CWERR (HCW_ERR_MEM)

  for (i = 0; i < rank; i++)				/* read data */
    start[i] = 0;
  HCHK (SDreaddata (sdsid, start, NULL, dim_sizes, data))

  *bp = data;						/* finish up */
  return (SUCCEED);

} /* CWreadbyte */

/**********************************************************************/

intn CWwritebyte (		/*** WRITE SATELLITE BYTE DATA ***/
  int32 sdsid,			/* HDF SD dataset id */
  uint8 *bp			/* byte array */
) {

char sds_name[MAX_NC_NAME];
int32 dim_sizes[MAX_VAR_DIMS];
int32 rank, data_type, n_attrs;
int32 i, dvals;
uint8 *data = bp;
int32 start[MAX_VAR_DIMS];

							/* check type */
  HCHK (SDgetinfo (sdsid, sds_name, &rank, dim_sizes, &data_type, &n_attrs))
  if (data_type != DFNT_UINT8)
    CWERR (HCW_ERR_CTYPE)
  for (i = 0, dvals = 1; i < rank; i++)
    dvals = dvals*dim_sizes[i];    

  for (i = 0; i < rank; i++)				/* write data */
    start[i] = 0;
  HCHK (SDwritedata (sdsid, start, NULL, dim_sizes, data))

  return (SUCCEED);					/* finish up */

} /* CWwritebyte */


/**********************************************************************/
