/**********************************************************************/
/*
     FILE: hdfinfo.c
  PURPOSE: To list various information from a CoastWatch HDF file.
   AUTHOR: Peter Hollemans
     DATE: 04/03/1999
  CHANGES: n/a

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1999, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hcwf.h"

/* defined constants */

#define UMESSAGE "\
usage: hcwinfo input\n\
where: input            The CoastWatch dataset to read.\n"
#define NARG		1
#define PROG		"hdfinfo"

#define DATE_FORMAT	"%Y/%m/%d"
#define TIME_FORMAT	"%H:%M:%S UTC"

/* defined macros */

							/* error checkers  */
#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 
#define CCHK(r,v) { \
  if ((r) == FAIL) \
    XIT (CWstrerror (), (v), -1) \
} 
#define HCHK(r,v) { \
  if ((r) == FAIL) \
    XIT (HEstring (HEvalue (1)), (v), -1) \
} 

/**********************************************************************/

void err (const char *m, const char *v) {
  fprintf (stderr, "%s: %s: %s\n", PROG, v, m);
} /* err */

/**********************************************************************/

void usage (int ecode) {
  fprintf (stderr, UMESSAGE);
  exit (ecode);
} /* usage */

/**********************************************************************/

intn readatt (int32 objid, char *att, VOIDP att_val) {

/* Reads an attribute. */

int32 att_index;
char att_name[MAX_NC_NAME];
int32 data_type, n_values;
char *string;

  if ((att_index = SDfindattr (objid, att)) == FAIL)
    return (FAIL);
  HCHK (SDattrinfo (objid, att_index, att_name, &data_type, &n_values), att)
  HCHK (SDreadattr (objid, att_index, att_val), att)
  if (data_type == DFNT_CHAR8) {
    string = att_val;
    string[n_values] = '\0';    
  } /* if */

  return (SUCCEED);

} /* readatt */

/**********************************************************************/

void globalinfo (int32 sdid) {

/* Prints the global information. */

char sat[20], sensor[20];
char pass_type[10], proj[60];
CWDATE date;
CWXFM xfm;
char date_str[20], time_str[20];
char origin[200];

  if (readatt (sdid, "satellite", sat) == FAIL)		/* get global atts */
    XIT ("attribute not found", "satellite", -1);
  readatt (sdid, "sensor", sensor);
  readatt (sdid, "pass_type", pass_type);
  readatt (sdid, "projection", proj);
  CCHK (CWgetdate (sdid, &date), "CWgetdate")
  CWdatestr (date_str, 20, DATE_FORMAT, 0, date);
  CWdatestr (time_str, 20, TIME_FORMAT, 0, date);
  CCHK (CWgetproj (sdid, &xfm), "CWgetproj")
  readatt (sdid, "origin", origin);

  printf ("  %-15s %s\n", "satellite", sat);		/* print global atts */
  printf ("  %-15s %s\n", "sensor", sensor);
  printf ("  %-15s %s\n", "pass_date", date_str);
  printf ("  %-15s %s\n", "start_time", time_str);
  printf ("  %-15s %s\n", "pass_type", pass_type);
  printf ("  %-15s %s\n", "projection", proj);
  printf ("  %-15s %-.10g %-.10g %-.10g %-.10g\n", "et_affine", 
    xfm->et_affine[0], xfm->et_affine[1], xfm->et_affine[2], 
    xfm->et_affine[3]);
  printf ("  %-15s %-.10g %-.10g\n", "", xfm->et_affine[4],
    xfm->et_affine[5]);
  printf ("  %-15s %ld\n", "rows", xfm->rows);
  printf ("  %-15s %ld\n", "cols", xfm->cols);
  printf ("  %-15s %s\n", "origin", origin);

} /* globalinfo */

/**********************************************************************/

const char *dtstring (int32 data_type) {

/* Returns a string pointer to an HDF data type name. */

static char *dtstrings[] = {
  "CHAR8", "UCHAR8", "INT8", "UINT8",
  "INT16", "UINT16",
  "INT32", "UINT32", "FLOAT32",
  "FLOAT64", "UNKNOWN"
};

  switch (data_type) {
    case DFNT_CHAR8: return (dtstrings[0]);
    case DFNT_UCHAR8: return (dtstrings[1]);
    case DFNT_INT8: return (dtstrings[2]);
    case DFNT_UINT8: return (dtstrings[3]);
    case DFNT_INT16: return (dtstrings[4]);
    case DFNT_UINT16: return (dtstrings[5]);
    case DFNT_INT32: return (dtstrings[6]);
    case DFNT_UINT32: return (dtstrings[7]);
    case DFNT_FLOAT32: return (dtstrings[8]);
    case DFNT_FLOAT64: return (dtstrings[9]);
    default: return (dtstrings[10]);
  } /* switch */  

} /* dtstring */

/**********************************************************************/

void varinfo (int32 sdid, int32 sds_index) {

/* Prints variable information. */

int sdsid;
char sds_name[MAX_NC_NAME];
int32 rank, dim_sizes[MAX_VAR_DIMS];
int32 data_type, n_attrs, uncal_data_type;
float64 scale, offset, scale_err, offset_err;
int cal, good;
const char *data_type_str;
char units[20];
int16 percent_good;

							/* get var atts */
  HCHK (sdsid = SDselect (sdid, sds_index), "SDselect")
  HCHK (SDgetinfo (sdsid, sds_name, &rank, dim_sizes, &data_type,
    &n_attrs), "SDgetinfo")
  data_type_str = dtstring (data_type);
  cal = (SDgetcal (sdsid, &scale, &scale_err, &offset, &offset_err,
    &uncal_data_type) != FAIL);
  if (readatt (sdsid, "units", units) == FAIL)
    strcpy (units, "-");
  good = (readatt (sdsid, "percent_good", &percent_good) != FAIL);

  if (sds_index == 0) {					/* print header */
    printf ("  %-15s %-10s %-15s %-12s %-12s %-6s\n", "Variable", "Type", 
      "Units", "Scale", "Offset", "% Good");  
  } /* if */

							/* print var atts */
  printf ("  %-15s %-10s %-15s ", sds_name, data_type_str, units);
  if (cal)
    printf ("%-12.10g %-12.10g ", scale, offset);
  else
    printf ("%-12s %-12s ", "-", "-");
  if (good)
    printf ("%-6hd\n", percent_good);
  else
    printf ("%-6s\n", "-");

  HCHK (SDendaccess (sdsid), sds_name);			/* finish up */

} /* varinfo */

/**********************************************************************/

int main (int argc, char *argv[]) {

char *ifname;
int32 sdid;
int32 n_datasets, n_file_attrs, i;

  argc--, argv++;					/* parse parameters */
  if (argc < NARG)
    usage (-1);
  ifname = argv[0];

  HCHK (sdid = SDstart (ifname, DFACC_READ), ifname)	/* open input */

  printf ("Contents of file %s:\n\n", ifname);		/* print global */
  printf ("Global information:\n");
  globalinfo (sdid);

  printf ("\nVariable information:\n");			/* print vars */
  HCHK (SDfileinfo (sdid, &n_datasets, &n_file_attrs), ifname)
  for (i = 0; i < n_datasets; i++) 
    varinfo (sdid, i);

  HCHK (SDend (sdid), ifname)
  return (0);

} /* main */    

/**********************************************************************/
