/**********************************************************************/
/*
     FILE: morph.c
  PURPOSE: To define a number of morphological transformation routines.
   AUTHOR: Peter Hollemans
     DATE: 1999/06/30
  CHANGES: 1999/10/12, PFH, added ; after done: label for IRIX cc
    NOTES: See the following paper for full details on the routines:

             Simpson, James J., "Image Masking Using Polygon Fills and
             "Morphological Transformations", REMOTE SENS. ENVIRON.
             40:161-183 (1992)

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "morph.h"

/* defined macros */

#define EXIT(m) {\
  fprintf (stderr, "\n%s: %s at line %d\n", __FILE__, (m), __LINE__);\
  exit (-1);\
}

/**********************************************************************/

void erode (			/*** PERFORM EROSION OPERATION ***/
  unsigned char *obj,		/* object to erode */
  size_t nl, size_t ns,		/* number of lines, samples in object */
  unsigned char *se,		/* structure element (0=off, 1=on) */
  size_t n,			/* structure element size (must be odd) */
  unsigned char back,		/* background pixel value */
  unsigned char inobj,		/* object pixel value */
  unsigned char iner		/* erosion pixel value */
) {

long i, j, si, sj, oi, oj, off, cen;
int legal;
size_t dvals;
unsigned char *dest_obj;

  /* Allocate and fill temporary destination object */
  /* ---------------------------------------------- */
  dvals = nl*ns;
  if ((dest_obj = malloc (dvals)) == NULL)
    EXIT ("out of memory")
  memcpy (dest_obj, obj, dvals);

  /* Find center of structure element */
  /* -------------------------------- */
  cen = (n+1)/2 - 1;

  /* Loop through all object pixels */
  /* ------------------------------ */
  for (i = 0; i < nl; i++)
    for (j = 0; j < ns; j++) {
      off = i*ns + j;    
      if (obj[off] == inobj) {

        /* Loop through all structure pixels */
        /* --------------------------------- */
        for (si = 0; si < n; si++)
          for (sj = 0; sj < n; sj++)
            if (se[si*n + sj]) {

              /* Determine if object pixel is in bounds */
              /* -------------------------------------- */
              oi = i + (si-cen);
              oj = j + (sj-cen);
              legal = (oi >= 0 && oi < nl && oj >= 0 && oj < ns);

              /* Erode object pixel */
              /* ------------------ */
              if (!legal || obj[oi*ns + oj] == back) {
                dest_obj[off] = iner;
                goto done;
              } /* if */
            
            } /* if */

      } /* if */
      done:
      ;
    } /* for */

  /* Copy data back into source object */
  /* --------------------------------- */
  memcpy (obj, dest_obj, dvals);
  free (dest_obj);

} /* erode */

/**********************************************************************/

void dilate (			/*** PERFORM DILATION OPERATION ***/
  unsigned char *obj,		/* object to dilate */
  size_t nl, size_t ns,		/* number of lines, samples in object */
  unsigned char *se,		/* structure element (0=off, 1=on) */
  size_t n,			/* structure element size (must be odd) */
  unsigned char back,		/* background pixel value */
  unsigned char inobj,		/* object pixel value */
  unsigned char indil		/* dilation pixel value */
) {

long i, j, si, sj, oi, oj, off, cen;
int legal;
size_t dvals;
unsigned char *dest_obj;

  /* Allocate and fill temporary destination object */
  /* ---------------------------------------------- */
  dvals = nl*ns;
  if ((dest_obj = malloc (dvals)) == NULL)
    EXIT ("out of memory")
  memcpy (dest_obj, obj, dvals);

  /* Find center of structure element */
  /* -------------------------------- */
  cen = (n+1)/2 - 1;

  /* Loop through all background pixels */
  /* ---------------------------------- */
  for (i = 0; i < nl; i++)
    for (j = 0; j < ns; j++) {
      off = i*ns + j;    
      if (obj[off] == back) {

        /* Loop through all structure pixels */
        /* --------------------------------- */
        for (si = 0; si < n; si++)
          for (sj = 0; sj < n; sj++)
            if (se[si*n + sj]) {

              /* Determine if object pixel is in bounds */
              /* -------------------------------------- */
              oi = i + (si-cen);
              oj = j + (sj-cen);
              legal = (oi >= 0 && oi < nl && oj >= 0 && oj < ns);

              /* Dilate object pixel */
              /* ------------------- */
              if (legal && obj[oi*ns + oj] == inobj) {
                dest_obj[off] = indil;
                goto done;
              } /* if */
            
            } /* if */

      } /* if */
      done:
      ;
    } /* for */

  /* Copy data back into source object */
  /* --------------------------------- */
  memcpy (obj, dest_obj, dvals);
  free (dest_obj);

} /* dilate */
  
/**********************************************************************/
