/**********************************************************************/
/*
     FILE: polyfill.c
  PURPOSE: An implementation of a recursive polygon fill algorithm.
   AUTHOR: Peter Hollemans
     DATE: 1999/03/06
  CHANGES: 1999/06/28, PFH, updated with more comments
    NOTES: See the following paper for full details on the algorithm:

             Simpson, James J., "Image Masking Using Polygon Fills and
             "Morphological Transformations", REMOTE SENS. ENVIRON.
             40:161-183 (1992)

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>

#include "polyfill.h"
#include "stack.h"

/* defined macros */

#define IVAL(l) (image[ ((l)->y - 1)*ns + ((l)->x - 1) ])
#define IVAL_R(l,i,j) (image[ ((l)->y - 1 + (j))*ns + ((l)->x - 1 + (i)) ])
#define IVAL_XY(x,y) (image[ ((y) - 1)*ns + (x - 1) ])
#define ABOVE(l) (IVAL_R ((l), 0, -1))
#define BELOW(l) (IVAL_R ((l), 0, 1))
#define RIGHT(l) (IVAL_R ((l), 1, 0))
#define LEFT(l) (IVAL_R ((l), -1, 0))
#define TOGGLE(i) ((i) == inpoly ? outpoly : inpoly)
#define LEGAL(x,y) (x <= ns && x >= 1 && y <= nl && y >= 1)
unsigned char tmp_ival;
#define EDGE(i) ((tmp_ival = (i)) != inpoly && tmp_ival != outpoly && tmp_ival != back)

#define EXIT(m) {\
  fprintf (stderr, "\n%s: %s at line %d\n", __FILE__, (m), __LINE__);\
  exit (-1);\
}

/* type declarations */

typedef struct {
  long x, y;
  unsigned char output;
} location_struct;
typedef location_struct *location;

/**********************************************************************/

location copyloc (		/*** CREATE COPY OF LOCATION ***/
  location loc			/* location to copy */
) {

location new_location;

  /* Allocate new location */
  /* --------------------- */
  if ((new_location = malloc (sizeof (location_struct))) == NULL)
    EXIT ("out of memory")

  /* Copy location data */
  /* ------------------ */
  new_location->x = loc->x;
  new_location->y = loc->y;
  new_location->output = loc->output;
  return (new_location);

} /* copyloc */

/**********************************************************************/

location newloc (		/*** CREATE NEW LOCATION ***/
  long x, long y,		/* location coordinates */
  unsigned char output		/* output value at location */
) {

location new_location;

  /* Allocate new location */
  /* --------------------- */
  if ((new_location = malloc (sizeof (location_struct))) == NULL)
    EXIT ("out of memory")

  /* Copy location data */
  /* ------------------ */
  new_location->x = x;
  new_location->y = y;
  new_location->output = output;
  return (new_location);

} /* newloc */

/**********************************************************************/

void polyfill (			/*** RECURSIVELY FILL POLYGON REGION ***/
  long sx, long sy,		/* starting (x, y) location (1-relative) */
  unsigned char back,		/* background value */
  unsigned char inpoly,		/* inside polygon value */
  unsigned char outpoly,	/* outside polygon value */
  int recursive,		/* recursion flag, 0=no, 1=yes */
  unsigned char *image,		/* byte image to fill */
  long nl, long ns		/* number of lines, number of samples */
) {

stack fill_stack, recursive_stack, edge_stack;
int push_above, push_below;
location loc;
unsigned char ival;
unsigned char output;
long i, j, vals, x, y;
int offsets[] = {-1, 0, 1};

  /* Initialize stacks */
  /* ------------------ */
  stackinit (&fill_stack);
  stackinit (&recursive_stack);
  stackpush (&fill_stack, newloc (sx, sy, inpoly));

  do {

    while (!stackempty (fill_stack)) {
 
      /* Reset push flags */
      /* ---------------- */
      push_above = 1;
      push_below = 1;
  
      /* Pop x,y location from fill stack */
      /* -------------------------------- */
      loc = stackpop (&fill_stack);
      output = loc->output;  

      /* Move all the way left */
      /* --------------------- */
      while (loc->x > 1 && (ival = LEFT(loc)) == back)
        loc->x--;
  
      /* Check for a polygon edge (for recursion) */
      /* ---------------------------------------- */
      if (loc->x > 1 && recursive && EDGE (ival))
        stackpush (&recursive_stack, newloc (loc->x-1, loc->y, output));
  
      do {
  
        /* Mark the location */
        /* ----------------- */
        IVAL(loc) = output;
  
        /* Check above and push location if needed */
        /* --------------------------------------- */        
        if (loc->y > 1) {
          if ((ival = ABOVE(loc)) != back)
            push_above = 1;
          else if (push_above && ival == back) {
            stackpush (&fill_stack, newloc (loc->x, loc->y-1, output));
            push_above = 0;
          } /* else if */      
        } /* if */
  
        /* Check below and push location if needed */
        /* --------------------------------------- */
        if (loc->y < nl) {
          if ((ival = BELOW(loc)) != back)
            push_below = 1;
          else if (push_below && ival == back) {
            stackpush (&fill_stack, newloc (loc->x, loc->y+1, output));
            push_below = 0;
          } /* else if */
        } /* if */
  
        /* Move right */
        /* ---------- */
        loc->x++;
  
      } while (loc->x <= ns && IVAL(loc) == back);
  
      /* Check for a polygon edge again (for recursion) */
      /* ---------------------------------------------- */
      if (loc->x <= ns && recursive && EDGE (IVAL(loc)))
        stackpush (&recursive_stack, copyloc (loc));
  
      /* Free location */
      /* ------------- */
      free (loc);
  
    } /* while */

    if (!stackempty (recursive_stack)) {

      /* Add recursion location to fill stack */
      /* ------------------------------------ */
      loc = stackpop (&recursive_stack);
      output = TOGGLE (loc->output);
      stackpush (&fill_stack, newloc (loc->x, loc->y, output));

      /* Fill the polygon edge itself */
      /* ---------------------------- */
      stackinit (&edge_stack);
      stackpush (&edge_stack, loc);
      while (!stackempty (edge_stack)) {
        loc = stackpop (&edge_stack);        
        if (EDGE (IVAL(loc)))
          IVAL (loc) = output;
        for (i = 0; i < 3; i++) {
          for (j = 0; j < 3; j++) {
            x = loc->x + offsets[i];
            y = loc->y + offsets[j];
            if (LEGAL (x, y) && EDGE (IVAL_XY (x, y)))
              stackpush (&edge_stack, newloc (x, y, 0));
          } /* for */
        } /* for */
        free (loc);  
      } /* while */  

    } /* if */

  } while (!stackempty (fill_stack));  
  
  /* Finish filling outside of polygon */
  /* --------------------------------- */
  for (i = 0, vals = nl*ns; i < vals; i++)
    if (image[i] != inpoly)
      image[i] = outpoly;
  
} /* polyfill */
  
/**********************************************************************/
